/******************* Part of full PrhoLog (RhoLog for Prolog) *******************

  Prholog-full is a  package for strategic rule-based programming
  with individual, function, sequence, and context variables and
  membership constraints over them.

  library.pl, written by Temur Kutsia and Besik Dundua.
  The library of built-in strategies.
  Last edit: May 14, 2016.
 
  Tested for SWI-Prolog (7.2.3 and later).

  (c) RISC, Johannes Kepler University Linz, Austria

********************************************************************************/

%--------------------------------------------------------------------------
%   all_answers returns a hedge consisting of all answers computed by 
%   the given strategy for the given hedge. To distinguish from each other, 
%   each answer is wrapped in the 'ans' constructor. Can be seen as an
%   analog of findall for PrhoLog. Exactly one strategy must be provided.
%--------------------------------------------------------------------------

all_answers([Strategy], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
    Str_args),
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        Constraint,[],One_result],
    findall(One_answer,
        (Strategy_appication,
         instance([hdg_rholog_internal,seq_var(s_X)],One_result,
             [hdg_rholog_internal|One_answer_temp]),
         One_answer = [ans|One_answer_temp]
        ),
        L),
    id([], [hdg_rholog_internal|L], Rhs, Constraint, Acc, Result).



%--------------------------------------------------------------------------
%   choice chooses a strategy from the given strategies that successfully 
%   applies to the given input hedge under given constraints. Output_hedge 
%   matches the result. With backtracking one can get all possible results. 
%   If there is no strategy that succeeds, choice fails.
%--------------------------------------------------------------------------

choice([Strategy1|_], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy1, Str1_name, 
        Str1_args),
    Strategy_appication =.. 
        [Str1_name,Str1_args,Lhs,Rhs,Constraint,Acc,Result],
    Strategy_appication.
choice([_|Strategies], Lhs, Rhs, Constraint, Acc, Result) :-
    choice(Strategies, Lhs, Rhs, Constraint, Acc, Result).



%-------------------------------------------------------------------------
%   closure applies the given strategy to the input hedge under the given
%   constraints repeteadly as often as possible, one after the other. 
%   Each time, Output_hedge matches the result. Results of applications 
%   are returned one by one, as they are computed via backtracking. Exactly 
%   one strategy must be provided.
%-------------------------------------------------------------------------

closure([_], Lhs, Rhs, Constraint, Acc, Result) :-
    id([], Lhs, Rhs, Constraint, Acc, Result).
closure([Strategy], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        Constraint,[],Result1],
    Strategy_appication,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    closure([Strategy], Lhs1, Rhs, Constraint, Acc, Result).




%----------------------------------------------------------------------------
%   compose first applies the first strategy from the given strategies to the 
%   input hedge under the given constraints. To its result, the second 
%   strategy is applied and so on. Output_hedge matches the final result.
%   Compose fails if one of the strategies fails. All possible results can be 
%   obtained by backtracking. At least two strategies must be provided.
%----------------------------------------------------------------------------

compose([Strategy1, Strategy2], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy1, Str1_name, 
        Str1_args),
    Strategy_appication1 =.. 
        [Str1_name,Str1_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        [],[],Result1],
    Strategy_appication1,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    get_strategy_name_arguments_from_the_list_form(Strategy2, Str2_name, 
        Str2_args),
    Strategy_appication2 =..
        [Str2_name,Str2_args,Lhs1,Rhs,Constraint,Acc,Result],
    Strategy_appication2.
        
compose([Strategy1, Strategy2, Strategy3|Strategies], Lhs, Rhs, Constraint, 
    Acc, Result
) :-
    get_strategy_name_arguments_from_the_list_form(Strategy1, Str1_name, 
        Str1_args),
    Strategy_appication1 =..
        [Str1_name,Str1_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        [],[],Result1],
    Strategy_appication1,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    compose([Strategy2, Strategy3|Strategies], Lhs1, Rhs, Constraint, 
        Acc, Result).


%--------------------------------------------------------------------------
%   first_all tries to apply the first strategy from the given 
%   strategies to the input hedge (lhs) under given constraints. 
%   If this fails, it tries the next strategy and so on. 
%   When a strategy is found that succeeds, first_all returns all 
%   answers computed by it, via backtracking. If no strategy succeeds, 
%   first_all fails.
%--------------------------------------------------------------------------

first_all([Strategy1|Strategies], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy1, Str1_name, 
        Str1_args),
    Strategy_appication =.. 
        [Str1_name,Str1_args,Lhs,Rhs,Constraint,Acc,Result],
    (   Strategy_appication *->
        true
    ;   first_all(Strategies, Lhs, Rhs, Constraint, Acc, Result)
    ).


%-----------------------------------------------------------------------
%   first_one behaves similarly to first_all, with the difference that when
%   a strategy is found that succeeds to the input hedge, first_one returns 
%   only one answer computed by it.
%-----------------------------------------------------------------------

first_one([Strategy1|_], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy1, Str1_name, 
        Str1_args),
    Strategy_appication=..[Str1_name,Str1_args,Lhs,Rhs,Constraint,Acc,Result],
    Strategy_appication,
    !.
first_one([_|Strategies], Lhs, Rhs, Constraint, Acc, Result) :-
    first_one(Strategies, Lhs, Rhs, Constraint, Acc, Result).

 

%-----------------------------------------------------------------------
%   id succeeds if the output hedge matches the input hedge. 
%   Otherwise it fails. No argument strategies are needed.
%-----------------------------------------------------------------------

id([], Lhs, Rhs, Constraint, Acc, Result) :-
    solve([Rhs << Lhs|Constraint], Subst),
    compose(Acc, Subst, Result).
        
        

%------------------------------------------------------------------------------
%   Interactive mode. It starts interaction with the user, asking her to 
%   provide a strategy, which it then applies to the input hedge. The process 
%   is repeated further so that the output hedge of the previous application 
%   becomes the input for the new strategy privided by the user, and so on. 
%   The interactive process stops when the user types 'finish'. The result is
%   the input hedge that was there at the moment when 'finish' was typed. 
%   It fails when the user-provided strategy fails for the current input hedge.
%-------------------------------------------------------------------------------

interactive([], Lhs, Rhs, Constraint, Acc, Result) :-
    symbol_list_to_term(Lhs, Lhs_hedge),
    symbol_list_to_term(Constraint, Constraint_hedge),
    writef('The hedge to be transformed\n\t%w\nunder the constraint\n\t%w.\n',
        [Lhs_hedge,Constraint_hedge]),
    write('Type the strategy name (or \'finish\' to stop transformation): '),
    read(Strategy),
    nl,
    Strategy \= finish,
    !,
    get_strategy_name_and_arguments(Strategy, Str_name, Str_args),
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_X)], 
        Constraint,[],Result1],
    Strategy_appication,
    !,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    interactive([], Lhs1, Rhs, Constraint, Acc, Result).
interactive([], Lhs, Rhs, Constraint, Acc, Result) :-
    solve([Rhs << Lhs|Constraint], Subst),
    compose(Acc, Subst, Result).



%----------------------------------------------------------------------------
%   N-fold iteration applies the given strategy repeatedly to the input hedge 
%   N times under Constraints. Output_hedge matches the result. Fails if one 
%   of the applications fails.
%----------------------------------------------------------------------------

iterate([Strategy,1], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,Lhs,Rhs,Constraint,Acc,Result],
    Strategy_appication.
iterate([Strategy, N], Lhs, Rhs, Constraint, Acc, Result) :-
    N > 1,
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        Constraint,[],Result1],
    Strategy_appication,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    N1 is N-1,
    iterate([Strategy,N1], Lhs1, Rhs, Constraint, Acc, Result).


      
%--------------------------------------------------------------------------
%   map maps the given strategy to each term of the input hedge (under the 
%   given constraints). Output of the strategy for a input term can be a 
%   hedge (or, in particular, a single term). The result is the hedge 
%   composed from the outputs of the strategy for each term in the input 
%   hedge. It fails as soon as the application of the strategy to a term 
%   from the input hedge fails.
%
%   When the input hedge is empty, the output hedge should be empty as well.
%---------------------------------------------------------------------------

map(_, [hdg_rholog_internal], [hdg_rholog_internal, seq_var(X)], _, Acc, 
    Result
) :-
    compose(Acc, [seq_var(X) ---> eps], Result).

map([Strategy], Lhs, Rhs, Constraint, Acc, Result):-
    solve([[hdg_rholog_internal, ind_var(i_X), seq_var(s_X)] << Lhs], Subst1),
    instance(ind_var(i_X), Subst1, Lhs_1_instance),
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,[hdg_rholog_internal,Lhs_1_instance],
        [hdg_rholog_internal,seq_var(s_Z)],[],[],Result1],
    Strategy_appication,
    instance([hdg_rholog_internal,seq_var(s_X)], Subst1, Lhs_new_instance),
    map([Strategy], Lhs_new_instance, [hdg_rholog_internal, seq_var(s_Y)],
        Constraint, Acc, Result2),
    instance(seq_var(s_Z), Result1, seq(Rhs_1_instance)),
    append([hdg_rholog_internal|Rhs_1_instance], [seq_var(s_Y)], Rhs_intermed),
    instance(Rhs_intermed, Result2, Rhs_instance),
    solve([Rhs << Rhs_instance | Constraint], Subst2),
    compose(Acc, Subst2, Result).

       
%----------------------------------------------------------------------------
%   map_to_subhedges: Similar to map above, with the only difference that the
%   strategy is mapped to not necesarily single terms from the input hedge,
%   but to nonempty subhedges of it.
%
%   When the input hedge is empty, the output hedge should be empty as well.
%----------------------------------------------------------------------------

map_to_subhedges(_, [hdg_rholog_internal], [hdg_rholog_internal,seq_var(X)], _, 
    Acc, Result
) :-
    compose(Acc, [seq_var(X) ---> eps], Result).

map_to_subhedges([Strategy], Lhs, Rhs, Constraint, Acc, Result) :-
    solve([[hdg_rholog_internal, seq_var(s_1), seq_var(s_X)] << Lhs], Subst1),
    instance(seq_var(s_1), Subst1, seq(Lhs_1_instance)),
    Lhs_1_instance \= [],
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,[hdg_rholog_internal|Lhs_1_instance],
        [hdg_rholog_internal,seq_var(s_2)],[],[],Result1],
    Strategy_appication,
    instance([hdg_rholog_internal,seq_var(s_X)], Subst1, Lhs_new_instance),
    map_to_subhedges([Strategy], Lhs_new_instance, [hdg_rholog_internal,seq_var(s_Y)],
        Constraint, Acc, Result2),
    instance(seq_var(s_2), Result1, seq(Rhs_1_instance)),
    append([hdg_rholog_internal|Rhs_1_instance], [seq_var(s_Y)], Rhs_intermed),
    instance(Rhs_intermed, Result2, Rhs_instance),
    solve([Rhs << Rhs_instance | Constraint], Subst2),
    compose(Acc, Subst2, Result).


%--------------------------------------------------------------------------
%   nf computes the normal form of application of the given strategy to the 
%   input hedge under the given constraints. Normal form is computed by 
%   applying the strategy to the input hedge repeatedly as long as the 
%   application succeeds. Whevener it fails, the current input hedge is 
%   returned as output hedge. Exactly one strategy must be provided.
%--------------------------------------------------------------------------

nf([Strategy], Lhs, Rhs, Constraint, Acc, Result) :-
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
    Str_args),
    nf_rholog_internal(Str_name, Str_args, Lhs, Rhs, Constraint, Acc, Result).

nf_rholog_internal(Str_name, Str_args, Lhs, Rhs, Constraint, Acc, Result) :-
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_X)],
        Constraint,[],Result1],
    Strategy_appication,
    instance([hdg_rholog_internal,seq_var(s_X)], Result1, Lhs1),
    nf_rholog_internal(Str_name, Str_args, Lhs1, Rhs, Constraint, Acc, Result).
nf_rholog_internal(Str_name, Str_args, Lhs, Rhs, Constraint, Acc, Result) :-
    Strategy_appication =..
        [Str_name,Str_args,Lhs,[hdg_rholog_internal,seq_var(s_)],
        Constraint,_,_],
    \+(Strategy_appication),
    id([], Lhs, Rhs, Constraint, Acc, Result).


%--------------------------------------------------------------------------
%   rewrite rewrites the input term by the given strategy. Rewriting
%   means that the a subterm of the input term (not necessarily the entire 
%   input term) is transformed by the strategy. It fails if Strategy fails 
%   for all subterms of the input term. All answers can be computed by 
%   backtracking. Exactly one strategy must be provided. Input and output 
%   are single terms, not arbitrary hedges.
%--------------------------------------------------------------------------

rewrite([Strategy], Lhs, Rhs, Constraint, Acc, Result) :-
    solve([[hdg_rholog_internal,[ctx_var(cvX),ind_var(i_X)]] << Lhs], Subst1),
    instance(ind_var(i_X), Subst1, L_instance),
    get_strategy_name_arguments_from_the_list_form(Strategy, Str_name, 
        Str_args),
    Strategy_appication =..
        [Str_name,Str_args,[hdg_rholog_internal,L_instance],
        [hdg_rholog_internal,ind_var(i_Y)],[],[],Result1],
    Strategy_appication,
    instance(ind_var(i_Y), Result1, R_instance),
    instance([hdg_rholog_internal,[ctx_var(cvX),R_instance]], Subst1, 
        Rhs_instance),
    solve([Rhs << Rhs_instance|Constraint], Subst2),
    compose(Acc, Subst2, Result).
  




