/******************* Part of full PrhoLog (RhoLog for Prolog) *******************

  Prholog-full is a  package for strategic rule-based programming
  with individual, function, sequence, and context variables and
  membership constraints over them.

  parse.pl, written by Temur Kutsia and Besik Dundua.
  Parses Prholog files and queries.
  Last edit: September 22, 2021.
 
  Tested for SWI-Prolog (7.2.3 and later).

  (c) RISC, Johannes Kepler University Linz, Austria

********************************************************************************/



    /****************************************
    *       Parsing Rholog Syntax          *
    ****************************************/


%-------------------------------------------------------------------------------
%
%   check_rholog_clause_syntax(+clause)
%   Checks Rholog clause for syntactic correctness and well-modedness.
%   Prolog subgoals and clause are not checked. It is a users
%   responsibility to use Prolog subgoals in the body of Rholog
%   clauses in such a way that well-modedness is not destroyed.
%
%-------------------------------------------------------------------------------

check_rholog_clause_syntax(end_of_file) :-
    !.
check_rholog_clause_syntax(Strategy :: Lhs ==> Rhs) :-
    !,
    check_rholog_clause_syntax(Strategy :: Lhs ==> Rhs where []).
check_rholog_clause_syntax(Strategy :: Lhs ==> Rhs where Constraint) :-
    !,
    nonvariable_strategy(Strategy, body),
    collect_variables_and_hole_from_a_hedge(Lhs, Lhs_vars),
    collect_variables_and_hole_from_a_hedge(Strategy, Strategy_vars),
    append(Strategy_vars, Lhs_vars, Strategy_and_Lhs_vars),
    delete_variables(Strategy_and_Lhs_vars,
        [ind_var_anonymous, fun_var_anonymous,
        seq_var_anonymous, ctx_var_anonymous],
        Strategy_and_Lhs_vars_without_anonymous),
    collect_variables_and_hole_from_a_hedge(Rhs, Rhs_vars),
    union( Lhs_vars, Rhs_vars, Vars_temp),
    union( Vars_temp,Strategy_vars, Vars_all),
    (    not(member(hole, Vars_all)),
    !
    ;    throw_exception(hole_in_the_program, body)
    ),
    (    subset(Rhs_vars,Strategy_and_Lhs_vars_without_anonymous),
    !
    ;    throw_exception(non_well_modedness_3, body)
    ),
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, body)
    ),
    collect_constraint_variables(Constraint, Constraint_vars, body),
    (    subset(Constraint_vars, Lhs_vars),
    !
    ;    throw_exception([non_well_modedness_2,Strategy], body)
    ),
    well_moded_constraint(Constraint, [], Strategy, body),
    !.
check_rholog_clause_syntax((Strategy :: Lhs ==> Rhs :- Body)) :-
    !,
    check_rholog_clause_syntax((Strategy :: Lhs ==> Rhs where [] :- Body)).
check_rholog_clause_syntax(
    (Strategy :: Lhs ==> Rhs where Constraint :- Body)
) :-
    !,
    nonvariable_strategy(Strategy, body),
    collect_variables_and_hole_from_a_hedge(Strategy, Strategy_vars),
    collect_variables_and_hole_from_a_hedge(Lhs, Lhs_vars),
    collect_variables_and_hole_from_a_hedge(Rhs, Rhs_vars),
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, body)
    ),
    collect_constraint_variables(Constraint, Constraint_vars, body),
    union(Lhs_vars, Rhs_vars, Vars_temp),
    union(Vars_temp,Strategy_vars, Vars_all),
    (    not(member(hole, Vars_all)),
    !
    ;    throw_exception(hole_in_the_program, body)
    ),
    (    subset(Constraint_vars, Lhs_vars),
    !
    ;    throw_exception([non_well_modedness_2,Strategy], body)
    ),
    well_moded_constraint(Constraint, [], Strategy, body),
    append(Strategy_vars, Lhs_vars, Strategy_and_Lhs_vars),
    check_rholog_clause_body_or_query_syntax(Body, Strategy_and_Lhs_vars, body,
        Strategy_and_Lhs_vars_plus_vars_collected_from_the_rhss_in_the_body),
    delete_variables(
        Strategy_and_Lhs_vars_plus_vars_collected_from_the_rhss_in_the_body,
        [ind_var_anonymous, fun_var_anonymous,
        seq_var_anonymous, ctx_var_anonymous],
        Strategy_and_Lhs_vars_plus_vars_collected_from_the_rhss_in_the_body_minus_anonymous),
    (   not(member(hole, 
           Strategy_and_Lhs_vars_plus_vars_collected_from_the_rhss_in_the_body)),
        !
    ;   throw_exception(hole_in_the_program, body)
    ),
    (   subset(Rhs_vars, 
           Strategy_and_Lhs_vars_plus_vars_collected_from_the_rhss_in_the_body_minus_anonymous),
        !
    ;   throw_exception(non_well_modedness_1, body)
    ).
check_rholog_clause_syntax(_).


%-------------------------------------------------------------------------------
%
%    check_rholog_clause_body_or_query_syntax is an auxiliary predicate that 
%    checks the body of a Rholog clause for syntactic correctness. The input 
%    variables serve to check well-modedness of the clause whose body is 
%    processed by this predicate.
%
%-------------------------------------------------------------------------------


check_rholog_clause_body_or_query_syntax(
    Strategy :: Lhs ==> Rhs, Vars_in, Body_or_query, Vars_out
) :-
    !,
    check_rholog_clause_body_or_query_syntax(Strategy :: Lhs ==> Rhs where [], 
        Vars_in, Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    Strategy :: Lhs ==> Rhs where Constraint, Vars_in, Body_or_query, Vars_out
) :-
    !,
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, Body_or_query)
    ),
    process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query(
        Strategy, Lhs, Rhs, Constraint, Vars_in, _, Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    (Strategy :: Lhs ==> Rhs, Others), Vars_in, Body_or_query, Vars_out
) :-
    !,
    check_rholog_clause_body_or_query_syntax((Strategy :: Lhs ==> Rhs where [], 
        Others), Vars_in, Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    (Strategy :: Lhs ==> Rhs where Constraint, Others), Vars_in, Body_or_query, 
    Vars_out
) :-
    !,
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, Body_or_query)
    ),
    process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query(
        Strategy, Lhs, Rhs, Constraint, Vars_in, _, Body_or_query, Vars_in_new),
    check_rholog_clause_body_or_query_syntax(Others, Vars_in_new, 
        Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    Strategy :: Lhs =\=> Rhs, Vars_in, Body_or_query, Vars_out
) :-
    !,
    check_rholog_clause_body_or_query_syntax(Strategy :: Lhs =\=> Rhs where [], 
        Vars_in, Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    Strategy :: Lhs =\=> Rhs where Constraint, Vars_in, Body_or_query, Vars_out
) :-
    !,
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, Body_or_query)
    ),
    process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query(
        Strategy, Lhs, Rhs, Constraint, Vars_in, Rhs_vars, Body_or_query, 
        Vars_out),
    (    subset(Rhs_vars, [i_,f_,s_,c_|Vars_in]),
         !
    ;    throw_exception(non_well_modedness_4, Body_or_query)
    ).
check_rholog_clause_body_or_query_syntax(
    (Strategy :: Lhs =\=> Rhs, Others), Vars_in, Body_or_query, Vars_out
) :-
    !,
    check_rholog_clause_body_or_query_syntax((Strategy :: Lhs =\=> Rhs where [], 
        Others), Vars_in, Body_or_query, Vars_out).
check_rholog_clause_body_or_query_syntax(
    (Strategy :: Lhs =\=> Rhs where Constraint, Others), Vars_in, Body_or_query, 
    Vars_out
) :-
    !,
    (    is_list(Constraint),
    !
    ;   throw_exception(constraint_is_not_list, Body_or_query)
    ),
    process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query(
        Strategy, Lhs, Rhs, Constraint, Vars_in, Rhs_vars, Body_or_query, 
        Vars_in_new),
    (   subset(Rhs_vars, [i_,f_,s_,c_|Vars_in]),
        !
    ;   throw_exception(non_well_modedness_4, Body_or_query)
    ),
    check_rholog_clause_body_or_query_syntax(Others, Vars_in_new, Body_or_query, 
        Vars_out).
check_rholog_clause_body_or_query_syntax(
        (Prolog_subgoal,Others), Vars_in, Body_or_query, Vars_out
) :-
    !,
    collect_variables_and_hole_from_a_hedge(Prolog_subgoal, Vars),
    (   not(member(hole, Vars)),
        !
    ;   throw_exception(hole_in_the_program, Body_or_query)
    ),
    (   are_sequence_or_context_variables(Vars),
        !
    ;   throw_exception(variables_in_prolog_clauses, Body_or_query)
    ),
    collect_function_variables(Vars, [], Function_vars),
    (   subset(Function_vars, Vars_in),
        !
    ;   throw_exception(non_well_modedness_9, Body_or_query)
    ),
    union(Vars_in, Vars, Vars_in_new),
    check_rholog_clause_body_or_query_syntax(Others, Vars_in_new, Body_or_query, 
        Vars_out).
check_rholog_clause_body_or_query_syntax(
    Prolog_subgoal, Vars_in, Body_or_query, Vars_out
) :-
    collect_variables_and_hole_from_a_hedge(Prolog_subgoal, Vars),
    (   not(member(hole, Vars)),
        !
    ;   throw_exception(hole_in_the_program, Body_or_query)
    ),
    (   are_sequence_or_context_variables(Vars),
        !
    ;   throw_exception(variables_in_prolog_clauses, Body_or_query)
    ),
    collect_function_variables(Vars, [], Function_vars),
    (   subset(Function_vars, Vars_in),
        !
    ;   throw_exception(non_well_modedness_9, Body_or_query)
    ),
    union(Vars_in, Vars, Vars_out).

are_sequence_or_context_variables([]).
are_sequence_or_context_variables([H|T]) :-
    atom_codes(H, N),
    not((N=[115,95|_]; N=[99,95|_])),
    are_sequence_or_context_variables(T).

collect_function_variables([], Acc, Acc).
collect_function_variables([H|T], Acc, Out) :-
    atom_codes(H, N),
    N=[102,95|_],
    !,
    collect_function_variables(T, [H|Acc], Out).
collect_function_variables([_|T], Acc, Out):-
    collect_function_variables(T, Acc, Out).


%-------------------------------------------------------------------------------
%
%   Check Rholog query syntax (to be done)
%
%-------------------------------------------------------------------------------

check_rholog_query_syntax(Query) :-
    check_rholog_clause_body_or_query_syntax(Query, [], query, _).
     

%-------------------------------------------------------------------------
%   process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query
%-------------------------------------------------------------------------

process_variables_for_syntactic_check_in_a_rholog_clause_body_or_query(
    Strategy, Lhs, Rhs, Constraint, Vars_in, Rhs_vars, Body_or_query, Vars_out
) :-
    delete_variables(Vars_in,
        [ind_var_anonymous, fun_var_anonymous,
        seq_var_anonymous, ctx_var_anonymous],
        Vars_in_minus_anonymous),
    collect_variables_and_hole_from_a_hedge(Strategy, Strategy_vars),
    collect_variables_and_hole_from_a_hedge(Lhs, Lhs_vars),
    append(Strategy_vars, Lhs_vars, Strategy_and_Lhs_vars),
    (   not(member(hole, Strategy_and_Lhs_vars)),
        !
    ;   throw_exception(hole_in_the_program, Body_or_query)
    ),
    (   subset(Strategy_and_Lhs_vars, Vars_in_minus_anonymous),
        !
    ;   throw_exception(non_well_modedness_4, Body_or_query)
    ),
    collect_constraint_variables(Constraint, Constraint_vars, Body_or_query),
    collect_variables_and_hole_from_a_hedge(Rhs, Rhs_vars),
    subtract(Rhs_vars, Vars_in, Vars_1),
    union(Lhs_vars, Rhs_vars, Vars_temp),
    union(Vars_temp, Strategy_vars, Vars_all),
    (   not(member(hole, Vars_all)),
        !
    ;   throw_exception(hole_in_the_program, Body_or_query)
    ),
    (   subset(Constraint_vars, Vars_1),
        !
    ;   throw_exception(non_well_modedness_5, Body_or_query)
    ),
    well_moded_constraint(Constraint, [], Strategy, Body_or_query),
    union(Vars_in, Rhs_vars, Vars_out).


%-------------------------------------------------------------------------------
%  Checking a constraint for well-modedness. Basically the check is reduced
%  to checking cyclic dependencies between constrained variables, where
%  dependency is a transitive closure of the 'depends on' relation.
%  A (sequence or context) variable X depends on a (sequence or context)
%  variable Y in a constraint C if 'X in R' belongs to C and Y occurs in R.
%-------------------------------------------------------------------------------

well_moded_constraint([], Dependencies, Strategy, Body_or_query) :-
    (   cyclic(Dependencies),
        throw_exception([cyclic,Strategy], Body_or_query)
    )
    ;   true.
well_moded_constraint([X in R|T], Dependencies, Strategy, Body_or_query ):-
    collect_variables_and_hole_from_a_hedge(R, R_vars),
    delete_variables(R_vars,
        [ind_var_anonymous, fun_var_anonymous,
        seq_var_anonymous, ctx_var_anonymous, ind_var, fun_var],
        R_with_seq_and_ctx_vars_only),
    well_moded_constraint(T,[[X, R_with_seq_and_ctx_vars_only]|Dependencies],
        Strategy, Body_or_query).

cyclic(Dependency_list):-
    member([Node,_], Dependency_list),
    cyclic(Node, Dependency_list, []).
cyclic(Node, _, Trail) :-
    member(Node, Trail),
    !.
cyclic(Node, Dependency_list, Trail):-
    append(_,[[Node, Successors]|_], Dependency_list),
    member(Next_node, Successors),
    cyclic(Next_node, Dependency_list, [Node|Trail]).


%-------------------------------------------------------------------------------
%   delete_variables(Variable_list_in, Variable_list_out, Variable_kinds_list)
%   Deletes from Variable_list_in all the variables whose kinds
%   (individual, sequence, function, or context) are specified in the
%   Variable_kinds_list.
%-------------------------------------------------------------------------------

delete_variables([],_, []).
delete_variables([H|T], Var_kinds, L) :-
    apply_alternatives(Var_kinds, H),
    !,
    delete_variables(T, Var_kinds, L).
delete_variables([H|T], Var_kinds, [H|L]) :-
    delete_variables(T, Var_kinds, L).

apply_alternatives([], _) :-
    !,
    fail.
apply_alternatives([Var_kind|Other], Arg) :-
    Predicate =.. [Var_kind,Arg],
    \+(Predicate),
    !,
    apply_alternatives(Other, Arg).
apply_alternatives(_, _).


%------------------------------------------------------------------------------
%   Checks whether a given strategy is a nonvariable strategy
%------------------------------------------------------------------------------

nonvariable_strategy(Strategy, Body_or_query) :-
    atomic(Strategy),
    rholog_var(Strategy),
    !,
    throw_exception(strategy_variables_1, Body_or_query).
nonvariable_strategy(Strategy, Body_or_query) :-
    var(Strategy),
    !,
    throw_exception(strategy_variables_2, Body_or_query).
nonvariable_strategy(_, _).


%-------------------------------------------------------------------------------
%   Collecting variables and hole that occur in the given hedge
%-------------------------------------------------------------------------------

collect_variables_and_hole_from_a_hedge(Hedge, Variables) :-
    Hedge =.. [H|T],
    collect_variables_and_hole_from_a_hedge([H|T], [], Variables).
collect_variables_and_hole_from_a_hedge([], Acc_variables, Acc_variables) :-
    !.
collect_variables_and_hole_from_a_hedge([H|T], Acc_variables, Variables) :-
    atomic(H),
    rholog_var(H),
    !,
    union([H], Acc_variables, Acc_variables_new),
    collect_variables_and_hole_from_a_hedge(T, Acc_variables_new, Variables).
collect_variables_and_hole_from_a_hedge([hole|T], Acc_variables, Variables) :-
    !,
    union([hole], Acc_variables, Acc_variables_new),
    collect_variables_and_hole_from_a_hedge(T, Acc_variables_new, Variables).
collect_variables_and_hole_from_a_hedge([H|T], Acc_variables, Variables) :-
    compound(H),
    !,
    H =.. H_list,
    collect_variables_and_hole_from_a_hedge(H_list, Acc_variables, 
        Acc_variables_new),
    collect_variables_and_hole_from_a_hedge(T, Acc_variables_new, Variables).
collect_variables_and_hole_from_a_hedge([_|T], Acc_variables, Variables) :-
    collect_variables_and_hole_from_a_hedge(T, Acc_variables, Variables).


%------------------------------------------------------------------------------
%   Collecting variables that occur in the given constraint
%------------------------------------------------------------------------------

collect_constraint_variables(Constraint, Constrained_vars, Body_or_query) :-
    collect_constraint_variables(Constraint, [], Constrained_vars, 
        Body_or_query).
collect_constraint_variables([], Acc_constrained, Acc_constrained, _) :-
    !.
collect_constraint_variables(
    [X in _|T], Acc_constrained, Constrained_vars, Body_or_query
) :-
    (   (
         \+(seq_var(X)),
         \+(ctx_var(X)),
         throw_exception(constrained_variables, Body_or_query)
        )
    ;   union(Acc_constrained, [X], Acc_constrained_new),
        collect_constraint_variables(T, Acc_constrained_new, Constrained_vars, 
            Body_or_query)
    ).

%-------------------------------------------------------------------------------
%   Preventing tautological strategy definitions
%-------------------------------------------------------------------------------

check_strategy_definition_for_tautology(Strategy, Strategy):-
    writef('Error. Strategy %q is defined as %q', 
        [Strategy, (Strategy:=Strategy)]),
    !,
    fail.

check_strategy_definition_for_tautology(_, _).

%-------------------------------------------------------------------------------
%   throw, catch and message hooks
%-------------------------------------------------------------------------------

throw_exception(Name, body) :-
    prolog_load_context(file, File),
%   If you use SWI-Prolog version 7.1.22 or earlier, uncomment the line below
%   and comment the next one.
%   prolog_load_context(term_position, '$stream_position'(_,Line,_,_,_)),
    prolog_load_context(term_position, '$stream_position'(_,Line,_,_)),
    throw(error(syntax_error(Name), file(File,Line,_,_))).
throw_exception(Name, query) :-
    throw(error(syntax_error(Name), query)).

user:message_hook(
    error(syntax_error([file_name,Rholog_relative_file_name]),_), user_error, _
) :-
    !,
    Lines=[ '~w \'~w\' ~w'-['The Rholog file',
        Rholog_relative_file_name,
        'does not have the required extension \'rho\'.']],
    print_message_lines(user_error, 'ERROR: ', Lines).

user:message_hook(
        error(syntax_error(constraint_is_not_list),
        file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. A constraint in this clause not a list.'],
    print_message_lines(user_error, 'ERROR: ', Lines).


user:message_hook(
        error(syntax_error(non_well_modedness_1),
        file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w'-['\n\tThe right hand side of the head of a rule contains 
        either an anonymous variable or a variable'],
        '~w'-['\n\tthat occurs neither in the left hand side of the head 
        nor in the right hand side of any subgoal in the body.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error([non_well_modedness_2,Strategy]),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w \'~w\', ~w'-['\n\tIn the rule for', Strategy, 
        'a constrained variable does not occur in the left hand side.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_3),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w ~w'-['\n\tThe right hand side of the clause head contains either 
        an anonymous variable or a variable',
        '\n\tthat occurs neither in the left hand side nor in the strategy 
        arguments.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_4),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w ~w'-['\n\tIn the clause body, the left hand side or the strategy 
        of a subgoal contains a named variable',
        '\n\tthat occurs neither in the right hand side of any subgoals 
        before it, nor in the left hand side on in the strategy of the head 
        of the clause.']],
    print_message_lines(user_error, 'ERROR: ', Lines).

user:message_hook(
    error(syntax_error(constraint_is_not_list), query), user_error, _
) :-
    !,
    Lines=['Syntax error. A constraint in the query is not a list.'],
    print_message_lines(user_error, 'ERROR: ', Lines).

user:message_hook(
    error(syntax_error(non_well_modedness_4), query), user_error, _
) :-
    !,
    Lines=['Syntax error. The query is not well-moded.',
        '~w ~w'-['\n\tIn the query, the left hand side or the strategy of a 
        subgoal contains a named variable',
        '\n\tthat does not occur in the right hand side of any subgoals 
        before it.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_5), 
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w ~w'-['\n\tIn the clause body, a constrained variable either 
        does not occur in the right hand side of the same subgoal,',
        '\n\tor occurs in the right hand side of a subgoal before it.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_5), query), user_error, _
) :-
    !,
    Lines=['Syntax error. The query is not well-moded.',
        '~w ~w'-['\n\tIn the query, a constrained variable either does not 
        occur in the right hand side of the same subgoal,',
        '\n\tor occurs in the right hand side of a subgoal before it.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(strategy_variables_1),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error.',
        '\n\tIn the head of a clause, the strategy name can not be a 
        Rholog variable.'],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(strategy_variables_2), 
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error.',
        '\n\tIn the head of a clause, the strategy name can not be a 
        Prolog variable.'],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error([cyclic, Strategy]),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error.',
        '~w \'~w\''-['\n\tA constrained variable depends on itself 
        in a rule for', Strategy]],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(constrained_variables),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error.',
        '\n\tOnly sequence or context variables can be constrained.'],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(constrained_variables), query), user_error, _
) :-
    !,
    Lines=['Syntax error in the query.',
        '\n\tOnly sequence or context variables can be constrained.'],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(variables_in_prolog_clauses),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error: ',
        '~w'-['\n\tSequence and context variables can not occur in 
        Prolog predicates.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(variables_in_prolog_clauses),query), user_error, _
) :-
    !,
    Lines=['Syntax error in the query.',
        '\n\tSequence and context variables can not occur in 
        Prolog predicates.'],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(hole_in_the_program),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line], 'Syntax error. ',
        '~w'-['\n\tThe protected constant hole can appear neither in the 
        program nor in queries.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(hole_in_the_program), query), user_error, _
) :-
    !,
    Lines=['Syntax error.',
        '~w'-['\n\tThe protected constant hole can appear neither in the 
        program nor in queries.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_9),
    file(File, Line, _,_)), user_error, _
) :-
    !,
    Lines=['~w. Clause at line ~d: '-[File,Line],
        'Syntax error. The clause is not well-moded.',
        '~w ~w'-['\n\tIn the clause body, the prolog subgoal contains a 
        function variable that occurs neither in the right hand side of 
        any subgoals before it,',
        '\n\tnor in the left hand side or in the strategy of the head 
        of the clause.']],
    print_message_lines(user_error, 'ERROR: ', Lines).
user:message_hook(
    error(syntax_error(non_well_modedness_9), query), user_error, _
) :-
    !,
    Lines=['Syntax error. The query is not well-moded.',
        '~w ~w'-['\n\tIn the query, the prolog subgoal contains a 
        function variable',
        '\n\tthat does not occur in the right hand side of any 
        subgoals before it.']],
    print_message_lines(user_error, 'ERROR: ', Lines).

user:message_hook(error(syntax_error(reading_error), _), user_error, _) :-
    !.
user:message_hook(X, user_error, _) :-
    print_message(error, X).


