/******************* Part of full PrhoLog (RhoLog for Prolog) *******************

  Prholog-full is a  package for strategic rule-based programming
  with individual, function, sequence, and context variables and
  membership constraints over them.

  solve.pl, written by Temur Kutsia.
  Solves matching problems and membership constraints.
  Last edit: May 14, 2016.
 
  Tested for SWI-Prolog (7.2.3 and later).

  (c) RISC, Johannes Kepler University Linz, Austria

********************************************************************************/


    /**********************************
    *       Constraint Solving        *
    *                                 *
    **********************************/

%-------------------------------------------------------------------------------
%
%   solve(Constraint, Solution) accepts a regular well-moded context sequence
%   constraint Constraint. If Constraint is solvable, a solving substitution
%   is returned in Solution. All the answers can be obtained by backtracking.
%
%-------------------------------------------------------------------------------

solve(Constraint, Smaller_without_anonymous) :-
    asserta(numb_rholog_internal(1000)),
    match(Constraint, [], S),
    remove_bindings_for_temporary_variables(S, Smaller_without_anonymous),
    retractall(numb_rholog_internal).

%-------------------------------------------------------------------------------
%   Conversions from input format to internal form 
%-------------------------------------------------------------------------------

%-------------------------------------------------------------------------------
%   term_to_list transforms a term into its list representation.
%   Example: The term 
%      f_F(a,c_G(s_X,h(b)),i_Y) 
%   will be translated into
%      [f_F, a, [s_G, s_X, [h, b]], i_Y]
%-------------------------------------------------------------------------------
        
term_to_list(Var, Term) :-
    var(Var),
    term_to_atom(Var, Term),
    !.
term_to_list(Ctx_or_seq_or_fun_or_ind_var_or_constant, X) :-
    functor(Ctx_or_seq_or_fun_or_ind_var_or_constant, _, 0),
    !,
    transform_var_or_constant(Ctx_or_seq_or_fun_or_ind_var_or_constant, X).
term_to_list(sstar(T), sstar(List)):-
    !,
    term_to_list(T, List).
term_to_list(sconc(T1,T2), sconc(List1,List2)) :-
    !,
    term_to_list(T1, List1),
    term_to_list(T2, List2).
term_to_list(sor(T1,T2), sor(List1,List2)) :-
    !,
    term_to_list(T1, List1),
    term_to_list(T2, List2).
term_to_list(eps, eps) :-
    !.
term_to_list(cstar(T), cstar(List)) :-
    !,
    term_to_list(T, List).
term_to_list(cconc(T1,T2), cconc(List1,List2)) :-
    !,
    term_to_list(T1, List1),
    term_to_list(T2, List2).
term_to_list(cor(T1,T2), cor(List1,List2)) :-
    !,
    term_to_list(T1, List1),
    term_to_list(T2, List2).
term_to_list(Term, List) :-
    Term =.. PseudoList,
    pseudo_list_to_list(PseudoList, List).


%-------------------------------------------------------------------------------
%   Transforms variables into terms of the form ctx_var(Var), seq_var(Var),
%   fun_var(Var), or ind_var(Var). Leaves constants unchanged.
%-------------------------------------------------------------------------------

transform_var_or_constant(Ind_Var, ind_var(Ind_Var)) :-
    ind_var(Ind_Var),
    !.
transform_var_or_constant(Fun_Var, fun_var(Fun_Var)) :-
    fun_var(Fun_Var),
    !.
transform_var_or_constant(Seq_Var, seq_var(Seq_Var)) :-
    seq_var(Seq_Var),
    !.
transform_var_or_constant(Ctx_Var, ctx_var(Ctx_Var)) :-
    ctx_var(Ctx_Var),
    !.
transform_var_or_constant('.', list_rholog_internal) :-
    !.
transform_var_or_constant([], nil_rholog_internal) :-
    !.
transform_var_or_constant(Term, Term):-
    functor(Term, _, 0),
    !.


%-------------------------------------------------------------------------------
%   pseudo_list_to_list takes a list of symbols and terms and returns
%   a list of symbols and list representations of terms.
%-------------------------------------------------------------------------------

pseudo_list_to_list([], []).
pseudo_list_to_list([Term|Other_terms], [Term_list|Other_terms_list]) :-
    term_to_list(Term, Term_list0),
    flatten_comma(Term_list0, Term_list),
    pseudo_list_to_list(Other_terms, Other_terms_list).
flatten_comma([',',[','|T1],[','|T2]], Res) :-
    !,
    append(T1, T2, R),
    flatten_comma([','|R], Res).
flatten_comma([',',[','|T1],T2], Res) :-
    !,
    append(T1, T2, R),
    flatten_comma([','|R], Res).
flatten_comma([',',T1,[','|T2]], [',',T1|R]) :-
    !,
    flatten_comma([','|T2], [','|R]).
flatten_comma([',',T1,T2], [',',TR1,TR2]):-
    !,
    flatten_comma(T1, TR1),
    flatten_comma(T2, TR2).
flatten_comma([], []):-
    !.
flatten_comma([X|L], [X1|L1]):-
    !,
    flatten_comma(X, X1),
    flatten_comma(L, L1).
flatten_comma(X, X).


%-------------------------------------------------------------------------------
%   term_constraints_to_symbol_list_constraints transforms
%   terms in constraints into their symbol list representations
%-------------------------------------------------------------------------------

term_constraints_to_symbol_list_constraints([], []).
term_constraints_to_symbol_list_constraints([S << T|Rest], [SL << TL|RestL]) :-
    term_to_list(S, SL),
    term_to_list(T, TL),
    term_constraints_to_symbol_list_constraints(Rest, RestL).
term_constraints_to_symbol_list_constraints([[Var,Regexp,Integer_flag]|Rest],
    [[seq_var(Var),RegexpL,Integer_flag]|RestL]
) :-
    seq_var(Var),
    !,
    regexp_to_list(Regexp, RegexpL),
    term_constraints_to_symbol_list_constraints(Rest, RestL).
term_constraints_to_symbol_list_constraints([[Var,Regexp,Integer_flag]|Rest],
    [[ctx_var(Var),RegexpL,Integer_flag]|RestL]
) :-
    ctx_var(Var),
    !,
    regexp_to_list(Regexp, RegexpL),
    term_constraints_to_symbol_list_constraints(Rest, RestL).
term_constraints_to_symbol_list_constraints([[Var,Regexp]|Rest], L) :-
    term_constraints_to_symbol_list_constraints([[Var,Regexp,0]|Rest], L).
term_constraints_to_symbol_list_constraints([Var in Regexp|Rest], L) :-
    term_constraints_to_symbol_list_constraints([[Var,Regexp,0]|Rest], L).


%-------------------------------------------------------------------------------
%   regexp_to_list transforms regular expressions
%   into their symbol list representations
%-------------------------------------------------------------------------------


regexp_to_list(sstar(R), sstar(RL)) :-
    !,
    regexp_to_list(R, RL).
regexp_to_list(sor(R1,R2), sor(R1L,R2L)) :-
    !,
    regexp_to_list(R1, R1L),
    regexp_to_list(R2, R2L).
regexp_to_list(sconc(R1,R2), sconc(R1L,R2L)) :-
    !,
    regexp_to_list(R1, R1L),
    regexp_to_list(R2, R2L).
regexp_to_list(eps, eps):-
    !.
regexp_to_list(cstar(R), cstar(RL)) :-
    !,
    regexp_to_list(R, RL).
regexp_to_list(cor(R1,R2), cor(R1L,R2L)) :-
    !,
    regexp_to_list(R1, R1L),
    regexp_to_list(R2, R2L).
regexp_to_list(cconc(R1,R2), cconc(R1L,R2L)) :-
    !,
    regexp_to_list(R1, R1L),
    regexp_to_list(R2, R2L).
regexp_to_list(Expr, Expr_list) :-
    term_to_list(Expr, Expr_list).


    /*******************************************************
    *  Conversions from internal form to the input format  *
    ********************************************************/

%-------------------------------------------------------------------------------
%   symbol_list_substitutions_to_term_substitutions transforms
%   symbol list representations of terms in substitutions
%   into their input format
%-------------------------------------------------------------------------------

symbol_list_substitutions_to_term_substitutions([], []) :-
    !.
symbol_list_substitutions_to_term_substitutions([X--->TL |Rest1], 
    [Y---> T|Rest2]
) :-
    symbol_list_to_term(X, Y),
    symbol_list_to_term(TL, T),
    symbol_list_substitutions_to_term_substitutions(Rest1, Rest2).

symbol_list_to_term(ctx_var(Var), Var) :-
    !.
symbol_list_to_term(fun_var(Var), Var) :-
    !.
symbol_list_to_term(seq_var(Var), Var) :-
    !.
symbol_list_to_term(ind_var(Var), Var) :-
    !.
symbol_list_to_term(nil_rholog_internal, []) :-
    !.
symbol_list_to_term(eps, eps) :-
    !.
symbol_list_to_term(seq([X1]), X2) :-
    !,
    symbol_list_to_term(X1, X2).
symbol_list_to_term(seq([H1,H2|Tail]), (X,Y)) :-
    !,
    symbol_list_to_term(H1, X),
    symbol_list_to_term(seq([H2|Tail]), Y).
symbol_list_to_term([hdg_rholog_internal,X1], X2) :-
    !,
    symbol_list_to_term(X1, X2).
symbol_list_to_term([hdg_rholog_internal,H1,H2|Tail], (X,Y)) :-
    !,
    symbol_list_to_term(H1, X),
    symbol_list_to_term([hdg_rholog_internal,H2|Tail], Y).
symbol_list_to_term([nil_rholog_internal|Tail], Term) :-
    !,
    symbol_list_to_term([[]|Tail], Term).
symbol_list_to_term([eps|Tail], Term) :-
    !,
    symbol_list_to_term(Tail, Term).
symbol_list_to_term([list_rholog_internal|Tail], Term) :-
    !,
    symbol_list_to_term(['.'|Tail], Term).
symbol_list_to_term([Head|Tail],Term) :-
    !,
    list_of_symbol_lists_to_symbol_list(Tail, T1),
    Term =.. [Head|T1].
symbol_list_to_term(Term, Term).
%symbol_list_to_term(Symbol, Term) :-
%   term_to_atom(Term, Symbol).       % 25.12            


list_of_symbol_lists_to_symbol_list([], []).
list_of_symbol_lists_to_symbol_list([X|T], ['.'|T1]) :-
    X == list_rholog_internal,
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], [[]|T1]) :-
    X == nil_rholog_internal,
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], T1) :-
    X == eps,
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], [X|T1]) :-
    \+ var(X),
    functor(X, _, 0),
    !,
%   term_to_atom(X1, X),                % 25.12
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], [X|T1]) :-
    var(X),
    !,
    list_of_symbol_lists_to_symbol_list(T, T1).
list_of_symbol_lists_to_symbol_list([X|T], [X1|T1]) :-
    symbol_list_to_term(X, X1),
    list_of_symbol_lists_to_symbol_list(T, T1).


    /*******************************************************
    *   Get substituitions that bind input variables only  *
    ********************************************************/


%-------------------------------------------------------------------------------
%   remove_bindings_for_temporary_variables gets a substitution as the input
%   and removes from it the bindings for all those variables that did not occur
%   in the input problem but have been generated at run-time.
%-------------------------------------------------------------------------------

remove_bindings_for_temporary_variables([], []).
remove_bindings_for_temporary_variables([ctx_var(Var) ---> _|Tail], New_subst) :-
    sub_atom(Var, _, _, _, temp_rholog_internal),
    !,
    remove_bindings_for_temporary_variables(Tail, New_subst).
remove_bindings_for_temporary_variables([fun_var(Var) ---> _|Tail], New_subst) :-
    sub_atom(Var, _, _, _, temp_rholog_internal),
    !,
    remove_bindings_for_temporary_variables(Tail, New_subst).
remove_bindings_for_temporary_variables([seq_var(Var) ---> _|Tail], New_subst) :-
    sub_atom(Var, _, _, _, temp_rholog_internal),
    !,
    remove_bindings_for_temporary_variables(Tail, New_subst).
remove_bindings_for_temporary_variables([ind_var(Var) ---> _|Tail], New_subst) :-
    sub_atom(Var, _, _, _, temp_rholog_internal),
    !,
    remove_bindings_for_temporary_variables(Tail, New_subst).
remove_bindings_for_temporary_variables([Var ---> T |Tail], [Var ---> T |Tail1]
) :-
    remove_bindings_for_temporary_variables(Tail, Tail1).


    /*******************************
    *    Transformation rules      *
    ******************************/

%-------------------------------------------------------------------------------
%   Basic transformation rules
%-------------------------------------------------------------------------------

% Success
match([], Acc, Acc).

% Trivial
match([X << X|Tail], Acc, Subst) :-
    !,
    match(Tail, Acc, Subst).

% Nullary funtions
match([[X|Xs] << X|Tail], Acc, Subst) :-
    !,
    match([[X|Xs] << [X]|Tail], Acc, Subst).
match([X << [X]|Tail], Acc, Subst) :-
    !,
    match(Tail,Acc,Subst).


%-------------------------------------------------------------------------------
%   Transformation rules for regular constraints for terms
%-------------------------------------------------------------------------------

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail], Acc, Subst) :-
    append(L1, [[seq_var(SeqVar),T,Flag]|L2], Tail),
    !,
    match_regexp_seq([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail], L1, 
        T, Flag, L2, Acc, Subst).


%-------------------------------------------------------------------------------
%   Transformation rules for regular expressions on contexts
%-------------------------------------------------------------------------------

match([[ctx_var(CtxVar),Arg] << T | Tail], Acc, Subst) :-
    append(L1, [[ctx_var(CtxVar),C,Flag]|L2], Tail),
    !,
    match_regexp_ctx([[ctx_var(CtxVar),Arg] << T | Tail], L1, 
        C, Flag, L2, Acc, Subst).


%-------------------------------------------------------------------------------
%   Rules that do not operate on regular expressions
%-------------------------------------------------------------------------------

% Replace an anonymous sequence variable with a fresh sequence variable
match([[X,s_|Xs] << [X|Ys]|Tail], Acc, Subst) :-
    !,
    anonymous_to_singleton_variables_in_term(s_, Singleton_seq_var),
    match([[X, Singleton_seq_var|Xs] << [X|Ys]|Tail], Acc, Subst).
        
% Sequence variable deletion
match([[X,seq_var(Seq_var)|Xs] << [X|Ys]|Tail], Acc, Subst) :-
    compose(Acc,[seq_var(Seq_var) ---> eps], New_acc),
    instance_of_term_list(Xs, [seq_var(Seq_var) ---> eps], Xs1),
    apply_substitution(Tail, [seq_var(Seq_var) ---> eps], Tail1),
    match([[X|Xs1] << [X|Ys]|Tail1], New_acc, Subst).

% Widening
match([[X,seq_var(Seq_var)|Xs] << [X,T1|Ys]|Tail], Acc, Subst) :-
    !,
    compose(Acc, [seq_var(Seq_var) ---> seq([T1,seq_var(Seq_var)])], 
        New_acc),
    instance_of_term_list(Xs, 
        [seq_var(Seq_var) ---> seq([T1,seq_var(Seq_var)])], Xs1),
    apply_substitution(Tail, 
        [seq_var(Seq_var) ---> seq([T1,seq_var(Seq_var)])], Tail1),
    match([[X,seq_var(Seq_var)|Xs1] << [X|Ys]|Tail1], New_acc, Subst).

% Replace an anonymous context variable with a fresh context variable
match([[c_,Arg] << T|Tail], Acc, Subst) :-
    anonymous_to_singleton_variables_in_term(c_, Singleton_ctx_var),
    match([[Singleton_ctx_var,Arg] << T|Tail], Acc, Subst).
        
% Context variable deletion
match([[ctx_var(Ctx_var),Arg] << T|Tail], Acc, Subst) :-
    \+(member([ctx_var(Ctx_var),_RegExp,_Flag], Tail)),
    compose(Acc, [ctx_var(Ctx_var) ---> hole], New_acc),
    instance(Arg, [ctx_var(Ctx_var) ---> hole], Arg1),
    apply_substitution(Tail, [ctx_var(Ctx_var) ---> hole], Tail1),
    match([Arg1 << T|Tail1], New_acc, Subst).

% Deepening
match([[ctx_var(Ctx_var),Arg] << [F|Args]|Tail], Acc, Subst):-
    \+(member([ctx_var(Ctx_var),_Reg_exp,_Flag], Tail)),
    append([F|L1], [Proj|L2], [F|Args]),
    append([F|L1], [[ctx_var(Ctx_var),hole]|L2], Context),
    compose(Acc, [ctx_var(Ctx_var) ---> Context], New_acc),
    instance(Arg, [ctx_var(Ctx_var) ---> Context], Arg1),
    apply_substitution(Tail, [ctx_var(Ctx_var) ---> Context], Tail1),
    match([[ctx_var(Ctx_var),Arg1] << Proj|Tail1], New_acc, Subst).

% Decomposition
match([[X,S1|Xs] << [X,T1|Ys]|Tail], Acc, Subst) :-
    !,
    match([S1 << T1,[X|Xs] << [X|Ys]|Tail], Acc, Subst).

% Function variable elimination
match([[f_|Xs] << [Y|Ys]|Tail], Acc, Subst) :-
    !,
    match([[Y|Xs] << [Y|Ys]|Tail], Acc, Subst).
match([[fun_var(X)|Xs] << [Y|Ys]|Tail], Acc, Subst) :-
    !,
    compose(Acc, [fun_var(X) ---> Y], New_acc),
    instance_of_term_list(Xs, [fun_var(X) ---> Y], Xs1),
    apply_substitution(Tail, [fun_var(X) ---> Y], Tail1),
    match([[Y|Xs1] << [Y|Ys]|Tail1], New_acc, Subst).
match([[f_|Xs] << T|Tail], Acc, Subst) :-
    T\=[_|_],
    T\=[],
    !,
    match([[T|Xs] << [T]|Tail], Acc, Subst).
match([f_ << T|Tail], Acc, Subst) :-
    match([[f_] << T|Tail], Acc, Subst).
match([[fun_var(X)|Xs] << T|Tail], Acc, Subst) :-
    T\=[_|_],
    T\=[],
    !,
    compose(Acc, [fun_var(X) ---> T], New_acc),
    instance_of_term_list(Xs, [fun_var(X) ---> T], Xs1),
    apply_substitution(Tail, [fun_var(X) ---> T], Tail1),
    match([[T|Xs1] << [T]|Tail1], New_acc, Subst).
match([fun_var(X) << T|Tail], Acc, Subst) :-
    match([[fun_var(X)] << T|Tail], Acc, Subst).

% Individual variable elimination
match([i_ << _Term|Tail], Acc, Subst) :-
    !,
    match(Tail, Acc, Subst).
match([ind_var(X) << Y|Tail], Acc, Subst) :-
    !,
    compose(Acc, [ind_var(X) ---> Y], New_acc),
    apply_substitution(Tail, [ind_var(X) ---> Y], Tail1),
    match(Tail1, New_acc, Subst).
match([[X,_,_]|Tail], _, _) :-
    \+(member(_ << _, Tail)),
    member([X,_,_], Tail),
    (X=seq_var(Y); X=ctx_var(Y)),
    write('ERROR. Variable '),
    write(Y),
    write(' constrained by several constraints 
        does not occur in equations.'),
    nl,
    abort.

% SAT
match([[_,_,0]|Tail], Acc, Subst) :-
    \+(member(_ << _, Tail)),
    match(Tail, Acc, Subst).
match([[X,_,1]|Tail], _, _) :-
    \+ member(_ << _, Tail),
    (X=seq_var(Y); X=ctx_var(Y)),
    write('ERROR. Constrained variable '),
    write(Y),
    write(', whose flag is 1, does not occur in equations.'),
    nl,
    abort.

% IRET  (Intersection of regular expressions for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, T1, 
    Flag1, L2, Acc, Subst
) :-
    append(L2_1, [[seq_var(Seq_var),T2,Flag2]|L2_2], L2),
    !,
    create_unique_temporary_seqvar(seq_var(Seq_var), Temp_seq_var),
    append(L1, [[seq_var(Seq_var),T1,Flag1]|L2_1], L3),
    append(L3, [[Temp_seq_var,T2,Flag2]|L2_2], New_tail),
    match([[X,seq_var(Seq_var)|Xs] << [X|Ys],
        [X,Temp_seq_var] << [X, seq_var(Seq_var)]|New_tail], Acc, Subst).

% VA  (Variable abstraction in regular expressions for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, T, 
    Flag, L2, Acc, Subst
) :-
    purify(T, PureT, New_constraints),
    !,
    append(L1, [[seq_var(Seq_var),PureT,Flag]|L2], L3),
    append(L3, New_constraints, New_tail),
    match([[X,seq_var(Seq_var)|Xs] << [X|Ys]|New_tail], Acc, Subst).

% ESRET (Empty sequence in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_],L1, eps, 
    Flag, L2, Acc, Subst
) :-
    !,
    evaluate_flag_for_seq_var(Flag, 0),
    append(L1, L2, L3),
    compose(Acc, [seq_var(Seq_var) ---> eps], New_acc),
    instance_of_term_list(Xs, [seq_var(Seq_var) ---> eps], New_Xs),
    apply_substitution(L3, [seq_var(Seq_var) ---> eps], New_tail),
    match([[X|New_Xs] << [X|Ys]|New_tail], New_acc, Subst).

% CRET (Concatenation in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, sconc(R1,R2), 
    Flag, L2, Acc, Subst
) :-
    !,
    evaluate_flag_for_seq_var(Flag, Value),
    create_unique_temporary_seqvar(seq_var(Seq_var), Temp_seq_var1),
    create_unique_temporary_seqvar(seq_var(Seq_var), Temp_seq_var2),
    new_flags_for_concatenation(Value,Temp_seq_var1, New_flag1, New_flag2),
    compose(Acc, [seq_var(Seq_var) ---> seq([Temp_seq_var1,Temp_seq_var2])], 
        New_acc),
    instance_of_term_list(Xs, 
        [seq_var(Seq_var) ---> seq([Temp_seq_var1,Temp_seq_var2])], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [seq_var(Seq_var) ---> seq([Temp_seq_var1,Temp_seq_var2])], Tail1),
    New_tail = 
        [[Temp_seq_var1,R1,New_flag1],[Temp_seq_var2,R2,New_flag2]|Tail1],
    match([[X,Temp_seq_var1,Temp_seq_var2|Xs1] << [X|Ys]|New_tail], 
        New_acc, Subst).

% ChRET (Choice in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_],L1, sor(R1,R2), 
    Flag, L2, Acc, Subst
) :-
    !,
    select_branch(sor(R1,R2), R_selected),
    evaluate_flag_for_seq_var(Flag, Value),
    append([[seq_var(Seq_var),R_selected,Value]|L1], L2, New_tail),
    match([[X,seq_var(Seq_var)|Xs] << [X|Ys]|New_tail], Acc, Subst).

% RRET1 (Repetition in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, sstar(_), 
    Flag, L2, Acc, Subst
) :-
    evaluate_flag_for_seq_var(Flag, 0),
    compose(Acc, [seq_var(Seq_var) ---> eps], New_acc),
    instance_of_term_list(Xs, [seq_var(Seq_var) ---> eps], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, [seq_var(Seq_var) ---> eps], New_tail),
    match([[X|Xs1] << [X|Ys]|New_tail], New_acc, Subst).

% RRET2 (Repetition in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, sstar(T), 
    _, L2, Acc, Subst
) :-
    !,
    create_unique_temporary_seqvar(seq_var(Seq_var), Temp_seq_var),
    compose(Acc,[seq_var(Seq_var) ---> seq([Temp_seq_var,seq_var(Seq_var)])], 
        New_acc),
    instance_of_term_list(Xs,
        [seq_var(Seq_var) ---> seq([Temp_seq_var,seq_var(Seq_var)])], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [seq_var(Seq_var) ---> seq([Temp_seq_var,seq_var(Seq_var)])], Tail1),
    New_tail = [[Temp_seq_var,T,1],[seq_var(Seq_var),sstar(T),0]|Tail1],
    match([[X,Temp_seq_var,seq_var(Seq_var)|Xs1] << [X|Ys]|New_tail], 
        New_acc, Subst).

% SVRET (Sequence variable in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, 
    seq_var(Another_seq_var), 0, L2, Acc, Subst
) :-
    !,
    compose(Acc, [seq_var(Seq_var) ---> seq_var(Another_seq_var)], New_acc),
    instance_of_term_list(Xs, 
        [seq_var(Seq_var) ---> seq_var(Another_seq_var)], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [seq_var(Seq_var) ---> seq_var(Another_seq_var)], New_tail),
    match([[X,seq_var(Another_seq_var)|Xs1] << [X|Ys]|New_tail], 
        New_acc, Subst).

match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, 
    seq_var(Another_seq_var), 1, L2, Acc, Subst
) :-
    !,
    create_unique_temporary_indvar(ind_var(Another_seq_var), 
        ind_var(Fresh_ind_var)),
    create_unique_temporary_seqvar(seq_var(Another_seq_var), 
        seq_var(Fresh_seq_var)),
    compose(Acc,[seq_var(Seq_var) ---> seq_var(Another_seq_var)], New_acc),
    instance_of_term_list(Xs, 
        [seq_var(Seq_var) ---> seq_var(Another_seq_var)], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, [seq_var(Seq_var) ---> seq_var(Another_seq_var)], 
        New_tail),
    match([[X,seq_var(Another_seq_var)|Xs1] << [X|Ys],
        [X,ind_var(Fresh_ind_var),
        seq_var(Fresh_seq_var)] << [X,seq_var(Another_seq_var)]|New_tail],
        New_acc, Subst).


% TRET (term in a regular expression for terms)
match_regexp_seq([[X,seq_var(Seq_var)|Xs] << [X|Ys]|_], L1, Term, 
    _, L2, Acc, Subst
) :-
    !,
    anonymous_to_singleton_variables_in_term(Term, New_term),
    compose(Acc, [seq_var(Seq_var) ---> New_term], New_acc),
    instance_of_term_list(Xs, [seq_var(Seq_var) ---> New_term], Xs1),
    append(L1, L2, L3),
    apply_substitution(L3, [seq_var(Seq_var) ---> New_term], New_tail),
    match([[X,New_term|Xs1] << [X|Ys]|New_tail], New_acc, Subst).

% IREC  (Intersection of regular expressions for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T | _], L1, C1, 
    Flag1, L2, Acc, Subst
) :-
    append(L2_1, [[ctx_var(Ctx_var),C2,Flag2]|L2_2], L2),
    !,
    create_unique_temporary_ctxvar(ctx_var(Ctx_var), Temp_ctx_var),
    append(L1, [[ctx_var(Ctx_var),C1,Flag1]|L2_1], L3),
    append(L3, [[Temp_ctx_var,C2,Flag2]|L2_2], New_tail),
    create_unique_temporary_constant(Const),
    match([[ctx_var(Ctx_var),Arg] << T, 
        [Temp_ctx_var,Const] << [ctx_var(Ctx_var),Const]|New_tail],
        Acc, Subst).

% VAC  (Variable abstraction in regular expressions for terms)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T |_], L1, C, 
    Flag, L2, Acc, Subst
) :-
    purify(C, PureC, New_constraints),
    !,
    append(L1, [[ctx_var(Ctx_var),PureC,Flag]|L2], L3),
    append(L3, New_constraints, New_tail),
    match([[ctx_var(Ctx_var),Arg] << T |New_tail], Acc, Subst).

% HREC (Hole in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, hole, 
    Flag, L2, Acc, Subst
) :-
    !,
    evaluate_flag_for_ctx_var(Flag, 0),
    append(L1, L2, L3),
    compose(Acc, [ctx_var(Ctx_var) ---> hole], New_acc),
    instance(Arg, [ctx_var(Ctx_var) ---> hole], New_arg),
    apply_substitution(L3, [ctx_var(Ctx_var) ---> hole], New_tail),
    match([New_arg << T|New_tail], New_acc, Subst).

% CREC (Concatenation in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, cconc(C1,C2), 
    Flag, L2, Acc, Subst
) :-
    !,
    evaluate_flag_for_ctx_var(Flag, Value),
    create_unique_temporary_ctxvar(ctx_var(Ctx_var), Temp_ctx_var1),
    create_unique_temporary_ctxvar(ctx_var(Ctx_var), Temp_ctx_var2),
    new_flags_for_concatenation(Value,Temp_ctx_var1, New_flag1, New_flag2),
    compose(Acc, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var1,[Temp_ctx_var2,hole]]], New_acc),
    instance(Arg, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var1,[Temp_ctx_var2,hole]]], New_arg),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var1,[Temp_ctx_var2,hole]]], Tail1),
    New_tail = 
        [[Temp_ctx_var1,C1,New_flag1],[Temp_ctx_var2,C2,New_flag2]|Tail1],
    match([[Temp_ctx_var1,[Temp_ctx_var2,New_arg]] << T|New_tail], 
        New_acc, Subst).

% ChREC (Choice in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, cor(C1,C2), 
    Flag, L2, Acc, Subst
) :-
    !,
    select_branch(cor(C1,C2), R_selected),
    evaluate_flag_for_ctx_var(Flag, Value),
    append([[ctx_var(Ctx_var),R_selected,Value]|L1], L2, New_tail),
    match([[ctx_var(Ctx_var),Arg] << T|New_tail], Acc, Subst).

% RREC1 (Repetition in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, cstar(_), 
    Flag, L2, Acc, Subst
) :-
    evaluate_flag_for_ctx_var(Flag, 0),
    compose(Acc, [ctx_var(Ctx_var) ---> hole], New_acc),
    instance(Arg, [ctx_var(Ctx_var) ---> hole], New_arg),
    append(L1, L2, L3),
    apply_substitution(L3, [ctx_var(Ctx_var) ---> hole], New_tail),
    match([New_arg << T|New_tail], New_acc, Subst).

% RREC2 (Repetition in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, cstar(R), 
    _, L2, Acc, Subst
) :-
    !,
    create_unique_temporary_ctxvar(ctx_var(Ctx_var), Temp_ctx_var),
    compose(Acc, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var,[ctx_var(Ctx_var),hole]]], 
        New_acc),
    instance(Arg, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var,[ctx_var(Ctx_var),hole]]], 
        New_arg),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [ctx_var(Ctx_var) ---> [Temp_ctx_var,[ctx_var(Ctx_var),hole]]], 
        Tail1),
    New_tail = [[Temp_ctx_var,R,1],[ctx_var(Ctx_var),cstar(R),0]|Tail1],
    match([[Temp_ctx_var,[ctx_var(Ctx_var),New_arg]] << T|New_tail], 
        New_acc, Subst).

% CVREC (Context variable in a regular expression for contexts)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, 
    [ctx_var(Another_ctx_var),hole], 0, L2, Acc, Subst
) :-
    !,
    compose(Acc, 
        [ctx_var(Ctx_var)--->[ctx_var(Another_ctx_var),hole]], New_acc),
    instance(Arg, 
        [ctx_var(Ctx_var)--->[ctx_var(Another_ctx_var),hole]], New_arg),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [ctx_var(Ctx_var) ---> [ctx_var(Another_ctx_var),hole]], New_tail),
    match([[ctx_var(Another_ctx_var),New_arg] << T|New_tail], 
        New_acc, Subst).

match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, 
    [ctx_var(Another_ctx_var),hole], 1, L2, Acc, Subst
) :-
    !,
    create_unique_temporary_funvar(fun_var(Another_ctx_var), 
        fun_var(Fun_var)),
    create_unique_temporary_seqvar(seq_var(Another_ctx_var), 
        seq_var(Seq_var1)),
    create_unique_temporary_ctxvar(ctx_var(Another_ctx_var), 
        ctx_var(Fresh_ctx_var)),
    create_unique_temporary_seqvar(seq_var(Another_ctx_var), 
        seq_var(Seq_var2)),
    create_unique_temporary_constant(Const),
    compose(Acc, 
        [ctx_var(Ctx_var)--->[ctx_var(Another_ctx_var),hole]], New_acc),
    instance(Arg, 
        [ctx_var(Ctx_var)--->[ctx_var(Another_ctx_var),hole]], New_arg),
    append(L1, L2, L3),
    apply_substitution(L3, 
        [ctx_var(Ctx_var) ---> [ctx_var(Another_ctx_var),hole]], New_tail),
    match([[ctx_var(Another_ctx_var),New_arg] << T, 
        [fun_var(Fun_var),seq_var(Seq_var1), [ctx_var(Fresh_ctx_var),Const], 
        seq_var(Seq_var2)] << [ctx_var(Another_ctx_var),Const]|New_tail],
        New_acc, Subst).

% CxRET1 (Context in a regular expression for contexts 1)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, 
    [ctx_var(Another_ctx_var),C], Flag, L2, Acc, Subst
) :-
    !,
    anonymous_to_singleton_variables_in_term(C, NewC),
    append(L1,[[ctx_var(Ctx_var), 
        cconc([ctx_var(Another_ctx_var),hole],NewC),Flag]|L2], New_tail),
    match([[ctx_var(Ctx_var),Arg] << T|New_tail], Acc, Subst).

% CxRET2 (Context in a regular expression for contexts 2)
match_regexp_ctx([[ctx_var(Ctx_var),Arg] << T|_], L1, Ctx, 
    _, L2, Acc, Subst
) :-
    !,
    anonymous_to_singleton_variables_in_term(Ctx, NewCtx),
    compose(Acc, [ctx_var(Ctx_var) ---> NewCtx], New_acc),
    apply_substitution([[ctx_var(Ctx_var),Arg] << T], 
        [ctx_var(Ctx_var) ---> NewCtx], NewEquation),
    append(L1, L2, L3),
    apply_substitution(L3, [ctx_var(Ctx_var) ---> NewCtx], New_tail),
    append(NewEquation, New_tail, NewProblem),
    match(NewProblem, New_acc, Subst).


        /*********************************************
        *        Operations on substitutions         *
        *********************************************/

%-------------------------------------------------------------------------------
%   compose(S1,S2,S) composes two substitutions S1 and S2 and returns S.
%-------------------------------------------------------------------------------

compose(S1, S2, S) :-
    apply_substitution(S1, S2, S3),
    remove_duplicated_vars_and_append_substs(S3, S2, S).


%-------------------------------------------------------------------------------
%   apply_substitution takes as input arguments the lists S1 and S2, and 
%   returns S3. 
%   S1 may contain bindings, matching equations, regular constraints, and terms. 
%   S2 is a substitution. 
%   S3 is obtained from S1 by applying S2 to terms in it 
%   (except variables X in regular constraints [X,R,F]).
%-------------------------------------------------------------------------------/

apply_substitution([], _, []):-
    !.
apply_substitution(S, [], S):-
    !.
apply_substitution([X ---> Y | Tail], S, [X ---> Y1 | Tail1]) :-
    !,
    instance(Y, S, Y1),
    apply_substitution(Tail, S, Tail1).
apply_substitution([X << Y | Tail], S, [X1 << Y1 | Tail1]) :-
    !,
    instance(X, S, X1),
    instance(Y, S, Y1),
    apply_substitution(Tail, S, Tail1).
apply_substitution([[X,Regexp,Flag]|Tail], S, [[X,Regexp1,Flag1]|Tail1]) :-
    !,
    instance(Regexp, S, Regexp1),
    instance(Flag, S, Flag1),
    apply_substitution(Tail, S, Tail1).
apply_substitution([X|Tail], S, [X1 |Tail1]) :-
    instance(X, S, X1),
    apply_substitution(Tail, S, Tail1).


%-------------------------------------------------------------------------------
%   instance(X,Subst,I) returns an instance I of X with respect to the
%   substitution Subst.
%-------------------------------------------------------------------------------

instance(circplus(X,Y), Subst, circplus(XInst,YInst)) :-
    !,
    instance(X, Subst, XInst),
    instance(Y, Subst, YInst).
instance(seq(L), Subst, Instance) :-
    instance_of_sequence(seq(L), Subst, Instance),
    !.
instance(seq_var(Seq_var), Subst, Instance) :-
    instance_of_sequence(seq_var(Seq_var), Subst, Instance),
    !.
instance(TermSequence_or_Regexp_or_Term, Subst, Instance):-
    instance_of_regexp(TermSequence_or_Regexp_or_Term, Subst, Instance),
    !.
instance(TermSequence_or_Regexp_or_Term, Subst, Instance) :-
    instance_of_nonseqvar_term(TermSequence_or_Regexp_or_Term, Subst, Instance).


% Instance of seq expressions.
instance_of_sequence(seq(L), Subst, seq(Instance)) :-
    instance_of_term_list(L, Subst, LInst),
    flatten_sequence(LInst, Instance).

instance_of_sequence(seq_var(Seq_var), Subst, Instance):-
    !,
    instance_of_symbol(seq_var(Seq_var), Subst, Instance).


% Instance of regular expressions.
instance_of_regexp(cstar(C), Subst, cstar(Instance)) :-
    !,
    instance_of_regexp(C, Subst, Instance).
instance_of_regexp(cor(C1,C2), Subst, cor(R1Inst,R2Inst)) :-
    !,
    instance_of_regexp(C1, Subst, R1Inst),
    instance_of_regexp(C2, Subst, R2Inst).
instance_of_regexp(cconc(C1,C2), Subst, cconc(R1Inst,R2Inst)) :-
    !,
    instance_of_regexp(C1, Subst, R1Inst),
    instance_of_regexp(C2, Subst, R2Inst).
instance_of_regexp(sstar(T), Subst, sstar(Instance)) :-
    !,
    instance_of_regexp(T, Subst, Instance).
instance_of_regexp(sor(T1,T2), Subst, sor(R1Inst,R2Inst)) :-
    !,
    instance_of_regexp(T1, Subst, R1Inst),
    instance_of_regexp(T2, Subst, R2Inst).
instance_of_regexp(sconc(T1,T2), Subst, sconc(R1Inst,R2Inst)) :-
    !,
    instance_of_regexp(T1, Subst, R1Inst),
    instance_of_regexp(T2, Subst, R2Inst).
instance_of_regexp(eps, _, eps) :-
    !.
instance_of_regexp(hole, _, hole) :-
    !.
instance_of_regexp([ctx_var(Ctx_var),hole], Subst, Context) :-
    !,
    instance_of_ctx_var(ctx_var(Ctx_var), Subst, Context).
instance_of_regexp(seq_var(Seq_var), Subst, Conc) :-
    !,
    instance_of_symbol(seq_var(Seq_var), Subst, Sequence),
    seq_to_conc(Sequence, Conc).
instance_of_regexp(Term, Subst, Instance) :-
    instance_of_nonseqvar_term(Term, Subst, Instance).

% Instance of terms that are not sequence variables
instance_of_nonseqvar_term([ctx_var(Ctx_var),Arg], Subst, Instance) :-
    !,
    instance_of_ctx_var(ctx_var(Ctx_var), Subst, Context),
    instance_of_nonseqvar_term(Arg, Subst, ArgInst),
    apply_context(Context, ArgInst, Instance).
instance_of_nonseqvar_term([Fun_var_or_FunSymb|ArgList], Subst, 
    [FunSymb|ArgInst]
) :-
    !,
    instance_of_symbol(Fun_var_or_FunSymb, Subst, FunSymb),
    instance_of_term_list(ArgList, Subst, ArgInst).
instance_of_nonseqvar_term(Symbol, Subst, Instance):-
    instance_of_symbol(Symbol, Subst, Instance).

% Instance of a list of terms.
instance_of_term_list([seq_var(Seq_var)|Other_terms], Subst, Instance) :-
    !,
    instance_of_symbol(seq_var(Seq_var), Subst, Sequence),
    instance_of_term_list(Other_terms, Subst, Other_termsInst),
    flatten_sequence([Sequence|Other_termsInst], Instance).
instance_of_term_list([Term|Other_terms], Subst, 
    [TermInstance|Other_termsInst]
) :-
    instance_of_nonseqvar_term(Term, Subst, TermInstance),
    instance_of_term_list(Other_terms, Subst, Other_termsInst).
instance_of_term_list([], _, []).


% Instance of a context variable.
instance_of_ctx_var(ctx_var(Ctx_var), Subst, Instance) :-
    member(ctx_var(Ctx_var)--->Instance, Subst),
    !.
instance_of_ctx_var(ctx_var(Ctx_var), _, [ctx_var(Ctx_var), hole]).

% Instance of a symbol.
instance_of_symbol(Var, Subst, Instance) :-
    member(Var--->Instance, Subst),
    !.
instance_of_symbol(Constant_or_var_not_in_dom, _, 
    Constant_or_var_not_in_dom).


% Transforming seq expressions into concatenation.
seq_to_conc(seq([X]), X) :-
    !.
seq_to_conc(seq([X,Y|T]), sconc(X,R)) :-
    seq_to_conc(seq([Y|T]), R).
seq_to_conc(eps, eps).


% Context application.
apply_context(hole, Term, Term) :-
    !.
apply_context([hole], Term, Term) :-
    !.
apply_context([Term1,hole], Term, [Term1,Term]) :-
    !.
apply_context([hole|Ctx_tail], Term, [Term|Ctx_tail]) :-
    !.
apply_context([[Ctx_head|Ctx_tail]|Tail], Term, [New_term|Tail]) :-
    apply_context([Ctx_head|Ctx_tail], Term, New_term).
apply_context([Head|Tail], Term, [Head|New_tail]) :-
    apply_context(Tail, Term, New_tail).

% Flattening nested seq expressions.
flatten_sequence([eps|Tail], Tail) :-
    !.
flatten_sequence([seq([])|Tail], Tail) :-
    !.
flatten_sequence([seq([X|T])|Tail], [X|Flat]) :-
    !,
    flatten_sequence([seq(T)|Tail], Flat).
flatten_sequence([X|Tail], [X1|Flat]) :-
    !,
    flatten_sequence(X, X1),
    flatten_sequence(Tail, Flat).
flatten_sequence(X, X).


%-------------------------------------------------------------------------------
%   remove_duplicated_vars_and_append_substs(S1,S2,S) takes two lists of 
%   bindings S1 and S2, deletes all the bindings X--->T from S2 such that 
%   S1 contains a binding for X, deletes all the bindings of the form X--->X 
%   from S1, and joins the obtained lists together into S.
%-------------------------------------------------------------------------------

remove_duplicated_vars_and_append_substs([], L, L) :-
    !.
remove_duplicated_vars_and_append_substs(L, [], L) :-
    !.
remove_duplicated_vars_and_append_substs(Subst1, Subst2, Result) :-
    remove_vars_from_second_that_occur_in_first(Subst1, Subst2, [], Subst2_new),
    remove_identities(Subst1, Subst1_new),
    append(Subst1_new, Subst2_new, Result).
remove_vars_from_second_that_occur_in_first([X--->_|Rest1], [X--->_|Rest2],
    ResultSoFar, Result
) :-
    !,
    remove_vars_from_second_that_occur_in_first(Rest1, Rest2, ResultSoFar, 
        Result).
remove_vars_from_second_that_occur_in_first([X--->Y|Rest1], [B|Rest2], Cleaned, 
    Result
) :-
    !,
    remove_vars_from_second_that_occur_in_first([X--->Y|Rest1], Rest2, 
        [B|Cleaned], Result).
remove_vars_from_second_that_occur_in_first([_|Rest1], [], Cleaned_of_first_var, 
    Result
) :-
    !,
    remove_vars_from_second_that_occur_in_first(Rest1, Cleaned_of_first_var, 
    [], Result).
remove_vars_from_second_that_occur_in_first([], Result, _, Result).

remove_identities([], []).
remove_identities([X--->X|Rest], Result):-
    !,
    remove_identities(Rest, Result).
remove_identities([B|Rest], [B|Result]):-
    remove_identities(Rest, Result).


    /*******************************
    *      Various utilities       *
    *******************************/

%-------------------------------------------------------------------------------
%   Generating unique temporary context, sequence, function and individual
%   variables.
%-------------------------------------------------------------------------------

create_unique_temporary_ctxvar(ctx_var(Ctx_var), ctx_var(Temp_ctx_var)) :-
    create_unique_temporary_symbol(Ctx_var, Temp_ctx_var).
create_unique_temporary_seqvar(seq_var(Seq_var), seq_var(Temp_seq_var)) :-
    create_unique_temporary_symbol(Seq_var, Temp_seq_var).
create_unique_temporary_funvar(fun_var(Fun_var), fun_var(Temp_fun_var)) :-
    create_unique_temporary_symbol(Fun_var, Temp_fun_var).
create_unique_temporary_indvar(ind_var(Ind_var), ind_var(Temp_ind_var)) :-
    create_unique_temporary_symbol(Ind_var, Temp_ind_var).
create_unique_temporary_constant(Const) :-
    create_unique_temporary_symbol(const, Const).
create_unique_temporary_symbol(Input, Unique) :-
    atom_concat(Input, temp_rholog_internal, Temp_name_base),
    my_gensym(Temp_name_base, Unique).

my_gensym(Base, Unique) :-
    numb_rholog_internal(N),
    atom_concat(Base, N, Unique),
    N1 is N + 1,
    retract(numb_rholog_internal(N)),
    asserta(numb_rholog_internal(N1)),
    !.


%-------------------------------------------------------------------------------
%   Flag manipulation in regular constraints: Generating new flags for the
%   constraints that arise after splitting concatenation, and evaluating
%   flag values.
%-------------------------------------------------------------------------------

% New flag generation
new_flags_for_concatenation(0, _, 0, 0).
new_flags_for_concatenation(1, Var, 0, circplus(Var,1)).

% Flag evaluation
evaluate_flag_for_seq_var(Flag, Flag) :-
    integer(Flag),
    !.
evaluate_flag_for_seq_var(circplus(X,Y), Value) :-
    nonempty_seq(X, V1),
    nonempty_seq(Y, V2),
    my_xor(circplus(V1,V2), Value).

evaluate_flag_for_ctx_var(Flag, Flag) :-
    integer(Flag),
    !.
evaluate_flag_for_ctx_var(circplus(X,Y), Value) :-
    nonempty_ctx(X, V1),
    nonempty_ctx(Y, V2),
    my_xor(circplus(V1,V2), Value).
    

my_xor(circplus(0,0), 1) :-
    !.
my_xor(circplus(0,1), 0) :-
    !.
my_xor(circplus(1,0), 0) :-
    !.
my_xor(circplus(1,1), 0) :-
    !.
my_xor(Unevaluated, Unevaluated).

nonempty_seq(eps, 0) :-
    !.
nonempty_seq(seq([]), 0) :-
    !.
nonempty_seq(S, 1) :-
    nonempty_seq_aux(S, 1),
    !.
nonempty_seq(S, S).

nonempty_seq_aux(seq([X|_]), 1) :-
    X \= seq_var(_),
    !.
nonempty_seq_aux(seq([_|Y]), V) :-
    nonempty_seq_aux(seq(Y), V).

nonempty_ctx(hole, 0).
nonempty_ctx(C, 1) :-
    nonempty_ctx_aux(C, 1),
    !.
nonempty_ctx(C, C).

nonempty_ctx_aux([[X|L]|_], 1) :-
    nonempty_ctx_aux([X|L], 1),
    !.
nonempty_ctx_aux([[_|_]|Tail], V) :-
    !,
    nonempty_ctx_aux(Tail, V).
nonempty_ctx_aux([H|_],1) :-
    H \= hole,
    H \= ctx_var(_),
    !.
nonempty_ctx_aux([_|Tail], V) :-
    nonempty_ctx_aux(Tail, V).


%-------------------------------------------------------------------------------
%   Selecting a branch in the 'sor' and 'cor' regular expression
%-------------------------------------------------------------------------------

select_branch(sor(R1,_R2), R1).
select_branch(sor(_R1,R2), R2).
select_branch(cor(R1,_R2), R1).
select_branch(cor(_R1,R2), R2).

%-------------------------------------------------------------------------------
%   anonymous_to_singleton_variables_in_subst replaces in a substitution each
%   anonymous variable with a new temporary singleton variable
%-------------------------------------------------------------------------------

anonymous_to_singleton_variables_in_subst([], []).
anonymous_to_singleton_variables_in_subst([X ---> RHS|Tail], 
    [X--->RHS1|Tail1]
) :-
    anonymous_to_singleton_variables_in_term(RHS, RHS1),
    anonymous_to_singleton_variables_in_subst(Tail, Tail1).

anonymous_to_singleton_variables_in_term(c_, Var) :-
    !,
    create_unique_temporary_ctxvar(ctx_var(c_An), Var).
anonymous_to_singleton_variables_in_term(s_,Var) :-
    !,
    create_unique_temporary_seqvar(seq_var(s_An), Var).
anonymous_to_singleton_variables_in_term(f_,Var) :-
    !,
    create_unique_temporary_funvar(fun_var(f_An), Var).
anonymous_to_singleton_variables_in_term(i_, Var) :-
    !,
    create_unique_temporary_indvar(ind_var(i_An), Var).
anonymous_to_singleton_variables_in_term(seq(L), seq(L1)) :-
    !,
    anonymous_to_singleton_variables_in_term(L, L1).
anonymous_to_singleton_variables_in_term([X], [Term]) :-
    !,
    anonymous_to_singleton_variables_in_term(X, Term).
anonymous_to_singleton_variables_in_term([X|Tail], [X1|Tail1]) :-
    !,
    anonymous_to_singleton_variables_in_term(X, X1),
    anonymous_to_singleton_variables_in_term(Tail, Tail1).
anonymous_to_singleton_variables_in_term(X, X).


%-------------------------------------------------------------------------------
%   select_equations takes a matching problem and returns a list of all
%   matching equations that occur in the input, removing from it all regular
%   constraints.
%-------------------------------------------------------------------------------

select_equations([], []).
select_equations([X << Y|Tail], [X << Y |Tail1]) :-
    !,
    select_equations(Tail, Tail1).
select_equations([_|Tail], L) :-
    select_equations(Tail, L).


%-------------------------------------------------------------------------------
%    Purification
%-------------------------------------------------------------------------------

purify(Ein, Eout, New_constraints):-
    purify(Ein, Eout, [], New_constraints),
    !,
    New_constraints \= [].
purify([], [], Acc, Acc) :-
    !.
purify([sstar(T)|Tail1], [Fresh_seq_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_seqvar(seq_var(s_), Fresh_seq_var),
    !,
    purify(Tail1,Tail2, [[Fresh_seq_var,sstar(T),0]|Acc], New_constraints).
purify([sconc(T1,T2)|Tail1], [Fresh_seq_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_seqvar(seq_var(s_), Fresh_seq_var),
    !,
    purify(Tail1,Tail2, [[Fresh_seq_var,sconc(T1,T2),0]|Acc], New_constraints).
purify([sor(T1,T2)|Tail1], [Fresh_seq_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_seqvar(seq_var(s_), Fresh_seq_var),
    !,
    purify(Tail1, Tail2, [[Fresh_seq_var,sor(T1,T2),0]|Acc], New_constraints).
purify([cstar(C)|Tail1], [Fresh_ctx_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_ctxvar(ctx_var(c_), Fresh_ctx_var),
    !,
    purify(Tail1, Tail2, [[Fresh_ctx_var,cstar(C),0]|Acc], New_constraints).
purify([cconc(C1,C2)|Tail1], [Fresh_ctx_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_ctxvar(ctx_var(c_), Fresh_ctx_var),
    !,
    purify(Tail1, Tail2, [[Fresh_ctx_var,cconc(C1,C2),0]|Acc], New_constraints).
purify([cor(C1,C2)|Tail1], [Fresh_ctx_var|Tail2], Acc, New_constraints) :-
    create_unique_temporary_ctxvar(ctx_var(c_), Fresh_ctx_var),
    !,
    purify(Tail1, Tail2, [[Fresh_ctx_var,cor(C1,C2),0]|Acc], New_constraints).
purify([[app,Ctx,Expr]|Tail1], [[Fresh_ctx_var,Expr1]|Tail2], Acc, 
    New_constraints
) :-
    create_unique_temporary_ctxvar(ctx_var(c_), Fresh_ctx_var),
    !,
    purify([Expr], [Expr1], [], Constr),
    append([[Fresh_ctx_var,Ctx,0]|Constr], Acc, New_acc),
    purify(Tail1, Tail2, New_acc, New_constraints).
purify([[H|T]|Tail1], [[H1|T1]|Tail2], Acc, New_constraints) :-
    purify([H|T], [H1|T1], [], Constr),
    append(Constr, Acc, New_acc),
    purify(Tail1, Tail2, New_acc, New_constraints).
purify([app,Ctx,Expr], [Fresh_ctx_var,Expr1], Acc, New_constraints) :-
    create_unique_temporary_ctxvar(ctx_var(c_), Fresh_ctx_var),
    !,
    purify([Expr], [Expr1], [[Fresh_ctx_var,Ctx,0]|Acc], New_constraints).
purify([X|Tail1], [X|Tail2], Acc, New_constraints):-
    purify(Tail1, Tail2, Acc, New_constraints).


    /********************************
    *       Recognizers             *
    ********************************/

%-------------------------------------------------------------------------------
%   Checks for NONANOMYMOUS context, sequence, function and individual
%   variables.
%-------------------------------------------------------------------------------


ind_var([]) :-
    !,
    fail.
ind_var([_|_]) :-
    !,
    fail.
ind_var(X) :-
    name(X, [105,95,_|_]).

fun_var([]) :-
    !,
    fail.
fun_var([_|_]) :-
    !,
    fail.
fun_var(X) :-
    name(X, [102,95,_|_]).

seq_var([]) :-
    !,
    fail.
seq_var([_|_]):-
    !,
    fail.
seq_var(X) :-
    name(X, [115,95,_|_]).

ctx_var([]) :-
    !,
    fail.
ctx_var([_|_]) :-
    !,
    fail.
ctx_var(X) :-
    name(X,[99,95,_|_]).


%-------------------------------------------------------------------------------
%      Recognizing anonymous variables
%-------------------------------------------------------------------------------

ind_var_anonymous(i_).
fun_var_anonymous(f_).
seq_var_anonymous(s_).
ctx_var_anonymous(c_).


%-------------------------------------------------------------------------------
%           Recognizing Rholog variables
%-------------------------------------------------------------------------------

rholog_var([]) :-
    !,
    fail.
rholog_var([_|_]) :-
    !,
    fail.
rholog_var(X) :-
    (    ind_var(X)
    ;    fun_var(X)
    ;    seq_var(X)
    ;    ctx_var(X)
    ;    X = i_
    ;    X = f_
    ;    X = s_
    ;    X = c_
    ),
    !.

