/******************************************************************************
  Name: trace.pl
  Description: The following code uses prolog_trace_interception to hook into 
    Prolog's trace algorithm. While tracing auxilary Prholog predicates are not 
    shown and hedges are converted from their internal representation to 
    increase readability. The code is split into 3 parts, disabling the trace
    for a single goal, for a goal including its subgoal and the part for 
    rewriting the output.
  Version: 1.0.0.0
  Author(s): Klaus Reisenberger-Hagmayer
  Created: 2015-11-17
  Last edited: 2015-11-17
  Last edited by: Klaus Reisenberger-Hagmayer
  Copyright:
  License:
  Notes:
    Note1: this section prevents return from being printed to output
    Note2: ~N inserts newline only if there is not already one inserted
    Note3: Lib predicates may use auxiliary predicates with the same
      functor. This auxiliary predicates should not be shown in the trace.
******************************************************************************/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Disables the tracer for a single goal. (The trace is still active for 
% subgoals)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%------------------------------------------------------------------------------
%% prolog trace interception(+Port, +Frame, +Choice, -Action)
%
% This predicate is the main entry point for hooking into the tracer. For a 
% detailed description see 'SWI-Prolog Reference Manual'.
%------------------------------------------------------------------------------

% Disable trace for special predicate rho_trace and notrace
prolog_trace_interception(_, Frame, _PC, skip) :-
    prolog_frame_attribute(Frame, goal, Goal),
    functor(Goal,Functor,_),
    member(Functor,[rho_trace, notrace]).

% Disables special predicates on the main Prholog process 
prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(
        Frame, goal, system:catch((check_rholog_query_syntax(_),
        compile_rholog_query(_, _)), _, print_message(user_error,_))
    ).

prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(
        Frame, goal, system:catch(user:(check_rholog_query_syntax(_),
        compile_rholog_query(_,_)), _, user:print_message(user_error,_))
    ).

prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(Frame, goal, compile_rholog_query(_,_)).

% Disables subpredicates of library-l.pl
prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(Frame, goal, nf_rholog_internal(_,_,_,_,_)).

prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(Frame, goal, rewrite_list(_,_,_,_)).

% Disables the step where the predicate is constructed by =..
prolog_trace_interception(_, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(Frame, goal,system: (_=..[Strategy|_])),
    prolog_frame_attribute(Frame, level, _),
    nb_getval(strategies, Strategies),
    nb_getval(lib_strategies, Lib_strategies),
    nb_getval(neg_strategies, Neg_strategies),
    (   member(Strategy, Strategies)
    ;   member(Strategy, Lib_strategies)
    ;   member(Strategy, Neg_strategies)
    ).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Disables the tracer for a goal including its subgoals.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Disables the tracer when one of the disabled_items is called and the tracer  
% is not already disabled
prolog_trace_interception(call, Frame, _PC, continue) :-
    use_modified_trace,
    nb_getval(is_prholog_trace_disabled, Is_disabled),
    not(Is_disabled),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,_),
    nb_getval(disabled_items, Disabled_list),
    member(Functor,Disabled_list),
    nb_setval(is_prholog_trace_disabled, true),
    nb_setval(current_predicate, Functor),
    nb_setval(current_level, Level),
    nb_getval(show_additional_trace_info, Show_info),
    (   Show_info ->    % DEBUG
        (nl, 
        write('disable Tracer('),
        write(call),
        write(','),
        write(Level),
        write('): '), 
        write(Functor),
        nl)
    ;   true
    ).

% Enables the tracer as soon as the same functor exits (the order of 
% predicates matters here, because of this, this part is not placed in the 
% next section)
prolog_trace_interception(exit, Frame, _PC, continue) :-
    use_modified_trace,
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,_),
    nb_getval(current_predicate,Current_predicate),
    nb_getval(current_level,Current_level),
    Functor = Current_predicate,
    Level = Current_level,
    nb_setval(is_prholog_trace_disabled, false),
    nb_setval(current_predicate, no_predicate),
    nb_setval(current_level, 0),
    nb_getval(show_additional_trace_info,Show_info),
    (   Show_info ->    % DEBUG
        (nl,
        write('enable Tracer('),
        write(exit),
        write(','),
        write(Level),
        write('): '),
        write(Functor),
        nl)
    ;   true
    ).

% If the trace is disabled and it fails then we have to enable it again. By 
% calling fail at the end we call the prolog_trace_interception of the 
% rewriting section for printing the output.
prolog_trace_interception(fail, Frame, _PC, continue) :-
    use_modified_trace,
    nb_getval(is_prholog_trace_disabled, Is_disabled),
    Is_disabled,
    nb_getval(last_called_predicate, Last_called_predicate),
    nb_getval(last_called_level,Last_called_level),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,_),
    Last_called_predicate = Functor,
    Last_called_level = Level,
    nb_setval(is_prholog_trace_disabled, false),
    nb_getval(show_additional_trace_info,Show_info),
    (   Show_info ->    % DEBUG
        (nl,
        write('strategy failed: ('),
        write(Level),
        write(') '),
        write(Goal),
        nl)
    ;   true
    ),
    fail. 
    
% If the trace is disabled and it fails at a deeper level than 
% Last_called_level a redo happens for Last_called_predicate. To handle
% this we have to enable the trace when this redo happens.
prolog_trace_interception(redo(_), Frame, _PC, continue) :-
    use_modified_trace,
    nb_getval(is_prholog_trace_disabled, Is_disabled),
    Is_disabled,
    nb_getval(last_called_predicate, Last_called_predicate),
    nb_getval(last_called_level,Last_called_level),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,_),
    Last_called_predicate = Functor,
    Last_called_level = Level,
    nb_setval(is_prholog_trace_disabled, false),
    nb_getval(show_additional_trace_info,Show_info),
    (   Show_info ->    % DEBUG
        (nl,
        write('redo happened: ('),
        write(Level),
        write(') '),
        write(Goal),
        nl)
    ;   true
    ),
    fail. 

% Skips all input while is_prholog_trace_disabled is true
prolog_trace_interception(Port, Frame, _PC, continue) :-
    use_modified_trace,
    nb_getval(is_prholog_trace_disabled, Is_disabled),
    Is_disabled,
    nb_getval(show_additional_trace_info,Show_info),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,_),
    (   Show_info ->    % DEBUG
        (nl,
        write('skipgoal('),
        write(Port),
        write(','),
        write(Level),
        write('): '),
        write(Functor),
        nl)
    ;   true
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Rewrites the printed goal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Rewrites output if Port equals call or redo
prolog_trace_interception(Port, Frame, _PC, Action) :-
    use_modified_trace,
    valid_ports_call_redo(Port,Trace_string),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,Arity),
    nb_getval(lib_strategies,Lib_strategies),
    nb_getval(strategies,Strategies),
    nb_getval(neg_strategies,Neg_strategies),
    (   member(Functor,Lib_strategies),
        Arity \= 4,
        Action = continue% see Note3
    ;   (   (   member(Functor,Strategies)
            ;   member(Functor,Lib_strategies)
            ),
            Strategy_type = '==>'
        ;   member(Functor,Neg_strategies),
            Strategy_type = '=\\=>'
        ),
        set_last_called(Port,Functor,Level),
        format(Trace_string, [Level]),
        to_strategy(Goal,Strategy_type),
        write(' ? '),
        flush_output,
        get_code(Response),
        (   Response = 10 -> 
            true
        ;   get_single_char(_)
        ),  % see [Note 1] above
        action(Response,Action)
    ).

% Rewrites output if Port equals exit or fail
prolog_trace_interception(Port, Frame, _PC, continue) :-
    use_modified_trace,
    valid_ports_exit_fail(Port,Trace_string),
    prolog_skip_level(_,very_deep),
    prolog_frame_attribute(Frame, goal, Goal),
    prolog_frame_attribute(Frame, level, Level),
    functor(Goal,Functor,Arity),
    nb_getval(lib_strategies,Lib_strategies),
    nb_getval(strategies,Strategies),
    nb_getval(neg_strategies,Neg_strategies),
    (   member(Functor,Lib_strategies),
        Arity \= 4 % see [Note 3]
    ;   (   (   member(Functor,Strategies)
            ;   member(Functor,Lib_strategies)
            ),
            Strategy_type = '==>'
        ;   member(Functor,Neg_strategies),
            Strategy_type = '=\\=>'
        ),
        format(Trace_string, [Level]),
        to_strategy(Goal,Strategy_type),
        format(' ? ~N') % see [Note 2] above
    ).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Global variables and auxiliary predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definition of global variables for the trace
:-nb_setval(is_modified_trace_enabled, false).
:-nb_setval(is_prholog_trace_disabled, false).
:-nb_setval(current_predicate, no_predicate).
:-nb_setval(current_level, 0).
% If set to true additional info is shown when the trace is activated
:-nb_setval(show_additional_trace_info, false).
% The next two lines are needed to handle the case when a strategy fails 
:-nb_setval(last_called_predicate, no_predicate).
:-nb_setval(last_called_level,0).
:-nb_setval(disabled_items, [
    check_rholog_query_syntax, compile_rholog_clause_body, 
    symbol_list_substitutions_to_term_substitutions, symbol_list_to_term, 
    % checkConstraints,
    get_strategy_name_arguments_from_the_applicative_form, 
    my_append, my_select,
    instance,
    get_strategy_name_arguments_from_the_list_form,
    solve,
    replace_rholog_ind_vars_by_prolog_vars,
    term_to_list,
    compose
    ]
).

% Sets the last called strategy and its level.
set_last_called(call,Functor,Level) :-
    nb_setval(last_called_predicate, Functor),
    nb_setval(last_called_level,Level).
set_last_called(_,_,_).

% Gets the output string for a given port
valid_ports_call_redo(call,'   Call: (~q) ').
valid_ports_call_redo(redo(_),'   Redo: (~q) ').

% Returns an action depending on the key pressed
action(115,skip). % s --> 115
action(97,abort). % a --> 97
action(_,continue).

% Gets the output string for a given port
valid_ports_exit_fail(exit,'   Exit: (~q) ').
valid_ports_exit_fail(fail,'   Fail: (~q) ').

to_strategy(Goal,Strategy_type) :-
    functor(Goal,Functor,Arity),
    arg(1,Goal,Strategy_arg),
    symbol_list_to_term(Strategy_arg, Strategy_term),
    arg(2,Goal,Input_hedge),
    symbol_list_to_term(Input_hedge, Input_term),
    arg(3,Goal,Output_hedge),
    symbol_list_to_term(Output_hedge, Output_term),
    (   Strategy_term = [] ->
        format('~q :: (~q) ', [Functor, Input_term])
    ;   format('~q(~q) :: (~q) ', [Functor, Strategy_term, Input_term])
    ),
    write(Strategy_type),
    format(' (~q)',[Output_term]),
    (   Arity > 3 ->
        arg(4,Goal,Constraints),
        (   Constraints = [] ->
            true
        ;   format(' where ~q',[Constraints]))
    ;  true
    ).

use_modified_trace :-
    nb_getval(is_modified_trace_enabled, Is_enabled),
    Is_enabled.
    

rho_trace :-
    nb_setval(is_modified_trace_enabled, true),
    trace.
    
notrace :-
    system:notrace,
    nb_setval(is_modified_trace_enabled, false).
